@@ -162,7 +162,7 @@ sub _is_selected_or_installed {
     my ($urpm, $db, $name) = @_;
 
     (grep { $_->flag_available } $urpm->packages_providing($name)) > 0 ||
-      $db->traverse_tag_find('name', $name, sub {}) > 0;
+      $db->traverse_tag('name', [ $name ], undef) > 0;
 }
 
 #- finds $pkg "provides" that matches $provide_name, and returns the version provided
@@ -350,7 +350,7 @@ sub _find_required_package__kmod {
     $choices->[0]->name =~ /^dkms-|-kernel-\d\./ or return;
 
     grep {
-	if (my ($version, $flavor, $release) = $_->name =~ /(:?.*)-kernel-(\d\..*)-(.*)-(.*)/) {
+	if (my ($version, $flavor, $release) = $_->name =~ /(?:.*)-kernel-(\d\..*)-(.*)-(.*)/) {
 	    my $kernel = "kernel-$flavor-$version-$release";
 	    _is_selected_or_installed($urpm, $db, $kernel);
 	} elsif ($_->name =~ /^dkms-/) {
@@ -983,7 +983,7 @@ sub resolve_requested__no_suggests_ {
 	    } else {
 		_set_flag_installed_and_upgrade_if_no_newer($db, $pkg);
 
-		if ($pkg->flag_installed && !$pkg->flag_upgrade) {
+		if ($pkg->flag_installed && !$pkg->flag_upgrade && !$urpm->{options}{downgrade}) {
 		    _no_more_recent_installed_and_providing($urpm, $db, $state, $pkg, $dep->{required}) or next;
 		}
 	    }
@@ -1188,9 +1188,9 @@ sub _unselect_package_deprecated_by_property {
 	    if ($o && $comparison > 0) {
 		#- installed package is newer
 		#- remove this package from the list of packages to install,
-		#- unless urpmi was invoked with --allow-force (in which
-		#- case rpm could be invoked with --oldpackage)
-		if (!$urpm->{options}{'allow-force'}) {
+		#- unless urpmi was invoked with --allow-force 
+		#- (in which case rpm could be invoked with --oldpackage)
+		if (!$urpm->{options}{'allow-force'} && !$urpm->{options}{downgrade}) {
 		    #- since the originally requested packages (or other
 		    #- non-installed ones) could be unselected by the following
 		    #- operation, remember them, to warn the user
